////////////////////////////////////////////////////////////////////////
/// \file  T962G4.cxx
/// \brief Use Geant4 to run the LArSoft detector simulation
///
/// \version $Id: T962G4.cxx,v 1.22 2010/07/20 06:08:30 bjpjones Exp $
/// \author  seligman@nevis.columbia.edu
////////////////////////////////////////////////////////////////////////
#ifndef T962G4_T962G4ANA_H
#define T962G4_T962G4ANA_H 

/// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Common/PtrVector.h"
#include "cetlib/exception.h"

// LArSoft includes
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nusimdata/SimulationBase/MCNeutrino.h"
#include "nutools/ParticleNavigation/ParticleList.h"
#include "lardataobj/Simulation/sim.h"
#include "larsim/MCCheater/ParticleInventoryService.h"
#include "larcore/Geometry/Geometry.h"

// art extensions
#include "nutools/RandomUtils/NuRandomService.h"

#include <cstring>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/stat.h>

#include <TTree.h>
#include "TH1.h"
#include "TSystem.h"
#include "CLHEP/Random/RandFlat.h"


///Geant4 interface 
namespace T962G4 {  
 
  class T962G4Ana : public art::EDAnalyzer{
  public:
   
    explicit T962G4Ana(fhicl::ParameterSet const& pset);
    virtual ~T962G4Ana();
    void analyze (const art::Event& evt); 
    void beginJob();

  private:
    TTree* ftree;
    std::string fG4ModuleLabel;  // name of the process module label that produced the input 
    std::string fGenieGenModuleLabel;
    std::string fpwd;
    int fm_run, fm_event, fm_trackID, fm_pdgcode, fm_neutrino_pdgcode; 
    float fm_mass, fm_energylost, fm_deflectionangle, fm_tpcexit_px, fm_tpcexit_py, fm_tpcexit_pz, fm_tpcexit_x, fm_tpcexit_y, fm_tpcexit_z, fm_tpcexit_energy, fm_minosenter_px, fm_minosenter_py, fm_minosenter_pz, fm_minosenter_x, fm_minosenter_y, fm_minosenter_z, fm_minosenter_energy, fm_init_px, fm_init_py, fm_init_pz, fm_init_x, fm_init_y, fm_init_z, fm_init_energy, fm_offset_x, fm_offset_y, fm_offset_z, fm_neutrino_px, fm_neutrino_py, fm_neutrino_pz, fm_neutrino_x, fm_neutrino_y, fm_neutrino_z, fm_neutrino_energy; 
  };


   static int event=1;
  //-----------------------------------------------------------------------
  // Constructor
  T962G4Ana::T962G4Ana(fhicl::ParameterSet const& pset) :
    EDAnalyzer(pset),
    fG4ModuleLabel(pset.get< std::string >("GeantModuleLabel")),
    fGenieGenModuleLabel             (pset.get< std::string >("GenieGenModuleLabel")),
    fm_run(0), 
    fm_event(0),
    fm_trackID(-1),
    fm_pdgcode(-1),
    fm_mass(-1.),
    fm_energylost(0.),
    fm_deflectionangle(0.),
    fm_tpcexit_px(0.),
    fm_tpcexit_py(0.),
    fm_tpcexit_pz(0.),
    fm_tpcexit_x(0.),
    fm_tpcexit_y(0.),
    fm_tpcexit_z(0.),
    fm_tpcexit_energy(0.),
    fm_minosenter_px(0.),
    fm_minosenter_py(0.),
    fm_minosenter_pz(0.),
    fm_minosenter_x(0.),
    fm_minosenter_y(0.),
    fm_minosenter_z(0.),
    fm_minosenter_energy(0.),
    fm_init_px(0.),
    fm_init_py(0.),
    fm_init_pz(0.),
    fm_init_x(0.),
    fm_init_y(0.),
    fm_init_z(0.),
    fm_init_energy(0.),
    fm_offset_x(0.),
    fm_offset_y(0.),
    fm_offset_z(0.)
  {
    // create a default random engine; obtain the random seed from NuRandomService,
    // unless overridden in configuration with key "Seed"
    art::ServiceHandle<rndm::NuRandomService>()
      ->createEngine(*this, pset, "Seed");


  }

  //-----------------------------------------------------------------------
  // Destructor
  T962G4Ana::~T962G4Ana() 
  {
  }

  //-----------------------------------------------------------------------
  void T962G4Ana::beginJob()
  {
  art::ServiceHandle<art::TFileService> tfs; 
  ftree= tfs->make<TTree>("MINOSinTree","MINOSinTree");
  ftree->Branch("run", &fm_run, "run/I");
  ftree->Branch("event", &fm_event, "event/I");
  ftree->Branch("trackID", &fm_trackID, "trackID/I");
  ftree->Branch("pdgcode", &fm_pdgcode, "pdgcode/I");
  ftree->Branch("mass", &fm_mass, "mass/F");
  ftree->Branch("energylost", &fm_energylost, "energylost/F");
  ftree->Branch("deflectionangle", &fm_deflectionangle, "deflectionangle/F");
  ftree->Branch("tpcexit_px", &fm_tpcexit_px, "tpcexit_px/F");
  ftree->Branch("tpcexit_py", &fm_tpcexit_py, "tpcexit_py/F");
  ftree->Branch("tpcexit_pz", &fm_tpcexit_pz, "tpcexit_pz/F");
  ftree->Branch("tpcexit_x", &fm_tpcexit_x, "tpcexit_x/F");
  ftree->Branch("tpcexit_y", &fm_tpcexit_y, "tpcexit_y/F");
  ftree->Branch("tpcexit_z", &fm_tpcexit_z, "tpcexit_z/F");
  ftree->Branch("tpcexit_energy", &fm_tpcexit_energy, "tpcexit_energy/F");
  ftree->Branch("minosenter_px", &fm_minosenter_px, "minosenter_px/F");
  ftree->Branch("minosenter_py", &fm_minosenter_py, "minosenter_py/F");
  ftree->Branch("minosenter_pz", &fm_minosenter_pz, "minosenter_pz/F");
  ftree->Branch("minosenter_x", &fm_minosenter_x, "minosenter_x/F");
  ftree->Branch("minosenter_y", &fm_minosenter_y, "minosenter_y/F");
  ftree->Branch("minosenter_z", &fm_minosenter_z, "minosenter_z/F");
  ftree->Branch("minosenter_energy", &fm_minosenter_energy, "minosenter_energy/F");  
  ftree->Branch("init_px", &fm_init_px, "init_px/F");
  ftree->Branch("init_py", &fm_init_py, "init_py/F");
  ftree->Branch("init_pz", &fm_init_pz, "init_pz/F");
  ftree->Branch("init_x", &fm_init_x, "init_x/F");
  ftree->Branch("init_y", &fm_init_y, "init_y/F");
  ftree->Branch("init_z", &fm_init_z, "init_z/F");
  ftree->Branch("init_energy", &fm_init_energy, "init_energy/F"); 
  ftree->Branch("neutrino_px", &fm_neutrino_px, "neutrino_px/F");
  ftree->Branch("neutrino_py", &fm_neutrino_py, "neutrino_py/F");
  ftree->Branch("neutrino_pz", &fm_neutrino_pz, "neutrino_pz/F");
  ftree->Branch("neutrino_x", &fm_neutrino_x, "neutrino_x/F");
  ftree->Branch("neutrino_y", &fm_neutrino_y, "neutrino_y/F");
  ftree->Branch("neutrino_z", &fm_neutrino_z, "neutrino_z/F");
  ftree->Branch("neutrino_energy", &fm_neutrino_energy, "neutrino_energy/F");
  ftree->Branch("neutrino_pdgcode", &fm_neutrino_pdgcode, "neutrino_pdgcode/I");
  ftree->Branch("offset_x", &fm_offset_x, "offset_x/F");
  ftree->Branch("offset_y", &fm_offset_y, "offset_y/F");
  ftree->Branch("offset_z", &fm_offset_z, "offset_z/F");

  fpwd = gSystem->pwd();
  std::cout << "T962G4Ana: pwd is: " << fpwd << '\n';
  }

  //-----------------------------------------------------------------------
void T962G4Ana::analyze(const art::Event& evt) 
{

  art::ServiceHandle<art::RandomNumberGenerator> rng;
  CLHEP::HepRandomEngine &engine = rng->getEngine();
  CLHEP::RandFlat   flat(engine);
  
  art::Handle< std::vector<simb::MCTruth> > mctruthListHandle;
  evt.getByLabel(fGenieGenModuleLabel,mctruthListHandle);


  art::PtrVector<simb::MCTruth> mclist;
  for (unsigned int ii = 0; ii <  mctruthListHandle->size(); ++ii)
    {
      art::Ptr<simb::MCTruth> mctparticle(mctruthListHandle,ii);
      mclist.push_back(mctparticle);
    } 

  art::Ptr<simb::MCTruth> mc(mclist[0]);
// 

//comment out this section for through going muon simulation

  if(mc->NeutrinoSet())
    {
      simb::MCNeutrino neut(mc->GetNeutrino());
      fm_neutrino_x =neut.Nu().Vx();
      fm_neutrino_y =neut.Nu().Vy();
      fm_neutrino_z =neut.Nu().Vz();
      fm_neutrino_px =neut.Nu().Px();
      fm_neutrino_py =neut.Nu().Py();
      fm_neutrino_pz =neut.Nu().Pz();
      fm_neutrino_energy=neut.Nu().E();
      fm_neutrino_pdgcode=neut.Nu().PdgCode();
    }
  //     
  
  
  //get the list of particles from this event
  art::ServiceHandle<cheat::ParticleInventoryService> pi_serv;
  const sim::ParticleList& plist = pi_serv->ParticleList();
  art::ServiceHandle<geo::Geometry> geom;

//  std::vector<const simb::MCParticle*> pvec;
//  for(unsigned int i = 0; i < plist.size(); ++i){
//    pvec.push_back(plist.Particle(i));
//  }

    fm_run=evt.id().run();
    fm_event=evt.id().event(); 
    fm_trackID=-1;
    fm_tpcexit_px=0.;
    fm_tpcexit_py=0.;
    fm_tpcexit_pz=0.;
    fm_tpcexit_x=0.;
    fm_tpcexit_y=0.;
    fm_tpcexit_z=0.;
    fm_tpcexit_energy=0.;
    fm_minosenter_energy=0.;
    fm_minosenter_px=0.;
    fm_minosenter_py=0.;
    fm_minosenter_pz=0.;
    fm_minosenter_x=0.;
    fm_minosenter_y=0.;
    fm_minosenter_z=0.;
    fm_pdgcode=0;
    fm_mass=0.;
    fm_energylost=0.;
    fm_deflectionangle=0.;
    fm_offset_x=117.3;//cm
    fm_offset_y=-19.3;//cm
    fm_offset_z=-152.95;//cm
    
    int numprimaries=0;
    int numinminos=0;
    int primindex=1;
    int priprimindex=0;
    sim::ParticleList::const_iterator itPart = plist.begin(),
      pend = plist.end(); // iterator to pairs (track id, particle)

    for(size_t iPart = 0; (iPart < plist.size()) && (itPart != pend); ++iPart){
      const simb::MCParticle* pPart = (itPart++)->second;
      if (!pPart) {
	throw art::Exception(art::errors::LogicError)
	  << "GEANT particle #" << iPart << " returned a null pointer";
      }
      
      if(pPart->Process()!="primary")
	continue; 
      numinminos++;
      simb::MCTrajectory trajectory = pPart->Trajectory();
      int numberofpoints= pPart->NumberTrajectoryPoints();    
      for(int j=1; j<numberofpoints; j++){
	TLorentzVector prevposition = trajectory.Position(j-1);
	TLorentzVector position = trajectory.Position(j);
	if(pPart->Process()=="primary"&&prevposition.Z()<152.95&&position.Z()>=152.95&&!(abs(pPart->PdgCode())==14||abs(pPart->PdgCode())==12))
	  numprimaries++;    	
      }    
    }
    //    std::ofstream myfile;
    std::ofstream myfile((fpwd+std::string("/MINOSin.txt")).c_str(), std::ios::out | std::ios::app);

    //myfile.open("MINOSin.txt", std::ios::out | std::ios::app);
    myfile << "________________________________________________________________________________\n";    
    myfile << "***** HEPEVT Common Event#: "<<event<<", "<<numprimaries+numinminos+1<<" particles (max 4000) ***** Double Precision\n";
    myfile << "4-byte integers, 8-byte floating point numbers, 4000-allocated entries.\n";
    myfile << "Indx Stat Par- chil-       (  P_x,       P_y,       P_z,    Energy,       M )\n";
    myfile << "      ID  ents dren    Prod (   X,         Y,         Z,        cT)      [mm]\n";
    myfile << "________________________________________________________________________________\n";
    myfile <<"   "<<1<<" +10001    "<<0<<"    0    (    "<<fm_neutrino_px<<",  "<<fm_neutrino_py<<",  "<<fm_neutrino_pz<<",  "<<fm_neutrino_energy<<",  "<<fm_mass<<")\n";
    if(fm_neutrino_pdgcode>0)
      myfile <<"      +"<<fm_neutrino_pdgcode<<"    0    0    (      "<<(fm_neutrino_x+fm_offset_x)*10.<<",       "<<(fm_neutrino_y+fm_offset_y)*10.<<",       "<<(fm_neutrino_z+fm_offset_z)*10.<<",  0.)\n";
    else
      myfile <<"      "<<fm_neutrino_pdgcode<<"    0    0    (      "<<(fm_neutrino_x+fm_offset_x)*10.<<",       "<<(fm_neutrino_y+fm_offset_y)*10.<<",       "<<(fm_neutrino_z+fm_offset_z)*10.<<",  0.)\n";

    
    itPart = plist.begin();
    pend = plist.end(); // iterator to pairs (track id, particle)

    for(size_t iPart = 0; (iPart < plist.size()) && (itPart != pend); ++iPart){
      const simb::MCParticle* pPart = (itPart++)->second;
      if (!pPart) {
	throw art::Exception(art::errors::LogicError)
	  << "GEANT particle #" << iPart << " returned a null pointer";
      }
      
      if(pPart->Process()!="primary")
	continue;
      
      simb::MCTrajectory trajectory = pPart->Trajectory();
      fm_pdgcode=pPart->PdgCode();       
      fm_mass=pPart->Mass();
      fm_trackID=iPart;
      TLorentzVector initposition = trajectory.Position(0);
      TLorentzVector initmomentum = trajectory.Momentum(0);
      int numberofpoints= pPart->NumberTrajectoryPoints();   
      
      fm_init_x=initposition.X();
      fm_init_y=initposition.Y();
      fm_init_z=initposition.Z();
      fm_init_px=initmomentum.Px();
      fm_init_py=initmomentum.Py();
      fm_init_pz=initmomentum.Pz();
      fm_init_energy=initmomentum.E(); 
      
      primindex++;
      
      myfile <<"   "<<primindex<<" +10001    1    0    (    "<<fm_init_px<<",  "<<fm_init_py<<",  "<<fm_init_pz<<",  "<<fm_init_energy<<",  "<<fm_mass<<")\n";
      if(fm_pdgcode>0)
      myfile <<"      +"<<fm_pdgcode<<"    1    0    (      "<<(fm_init_x+fm_offset_x)*10.<<",       "<<(fm_init_y+fm_offset_y)*10.<<",       "<<(fm_init_z+fm_offset_z)*10.<<",  2.64e+06)\n";
      else
      myfile <<"      "<<fm_pdgcode<<"    1    0    (      "<<(fm_init_x+fm_offset_x)*10.<<",       "<<(fm_init_y+fm_offset_y)*10.<<",       "<<(fm_init_z+fm_offset_z)*10.<<",  2.64e+06)\n";
    for(int j=1; j<numberofpoints; j++)
    {
      TLorentzVector prevposition = trajectory.Position(j-1);
      TLorentzVector position = trajectory.Position(j);
      TLorentzVector prevmomentum = trajectory.Momentum(j-1);
      TLorentzVector momentum = trajectory.Momentum(j);

       if(       (((prevposition.X()<0.&&position.X()>=0.)||(prevposition.X()>0.&&position.X()<=0.))&&(prevposition.Y()<20.&&prevposition.Y()>=-20.&&prevposition.Z()>0.&&prevposition.Z()<=90.))
       ||       (((prevposition.X()<47.&&position.X()>=47.)||(prevposition.X()>47.&&position.X()<=47.))&&(prevposition.Y()<20.&&prevposition.Y()>=-20.&&prevposition.Z()>0.&&prevposition.Z()<=90.))
       ||       (((prevposition.Y()<20.&&position.Y()>=20.)||(prevposition.Y()>20.&&position.Y()<=20.))&&(prevposition.X()<47.&&prevposition.X()>=0.&&prevposition.Z()>0.&&prevposition.Z()<=90.))
       ||       (((prevposition.Y()<-20.&&position.Y()>=-20.)||(prevposition.Y()>-20.&&position.Y()<=-20.))&&(prevposition.X()<47.&&prevposition.X()>=0.&&prevposition.Z()>0.&&prevposition.Z()<=90.))
       ||       (((prevposition.Z()<0.&&position.Z()>=0.)||(prevposition.Z()>0.&&position.Z()<=0.))&&(prevposition.Y()<20.&&prevposition.Y()>=-20.&&prevposition.X()>0.&&prevposition.X()<=47.))
       ||       (((prevposition.Z()<90.&&position.Z()>=90.)||(prevposition.Z()>90.&&position.Z()<=90.))&&(prevposition.Y()<20.&&prevposition.Y()>=-20.&&prevposition.X()>0.&&prevposition.X()<=47.))
       )
       {    
       fm_tpcexit_x=position.X();
       fm_tpcexit_y=position.Y();
       fm_tpcexit_z=position.Z();
       fm_tpcexit_px=momentum.Px();
       fm_tpcexit_py=momentum.Py();
       fm_tpcexit_pz=momentum.Pz();
       fm_tpcexit_energy=momentum.E(); 
       }
    // if(prevposition.Z()<154.22&&position.Z()>154.22&&abs(pPart->PdgCode())!=14&&abs(pPart->PdgCode())!=12)
//            {
       if(prevposition.Z()<152.95&&position.Z()>=152.95&&abs(pPart->PdgCode())!=14&&abs(pPart->PdgCode())!=12)
        {
       fm_minosenter_x=position.X();
       fm_minosenter_y=position.Y();
       fm_minosenter_z=position.Z();
       fm_minosenter_px=momentum.Px();
       fm_minosenter_py=momentum.Py();
       fm_minosenter_pz=momentum.Pz();
       fm_minosenter_energy=momentum.E();

       primindex++;
       priprimindex++;
       myfile <<"   "<<primindex<<"   +1    "<<primindex-1<<"    0    (    "<<fm_minosenter_px<<",  "<<fm_minosenter_py<<",  "<<fm_minosenter_pz<<",  "<<fm_minosenter_energy<<",  "<<fm_mass<<")\n";

	      //  2.64e+06 is 8 musec. I now add a random offset to this
	      // that is as large as the 10 musec NuMI beam width. Want
	      // to only do this if I've seen a 10001 and a 1 and I'm now
	      // on a +/- 13. That should only smear in time the overlay
	      // snarls. EC, 30-Jan-2013.
	      double tauNom(2.64e+06); // mm
	      double tauRan(10.0*flat.fire(1.0)); // musec
	      double tau(tauNom);

	      if (priprimindex>1 && abs(fm_pdgcode)==13)
		tau = tauNom + (tauRan*2.99792E11*1.e-6);

	      if(fm_pdgcode>0)
		//    myfile <<"       +"<<fm_pdgcode<<"    "<<primindex-1<<"    0    (      "<<(fm_minosenter_x+fm_offset_x)*10.<<",       "<<(fm_minosenter_y+fm_offset_y)*10.<<",       "<<(fm_minosenter_z+fm_offset_z)*10.<<",  2.64e+06)\n";
		myfile <<"       +"<<fm_pdgcode<<"    "<<primindex-1<<"    0    (      "<<(fm_minosenter_x+fm_offset_x)*10.<<",       "<<(fm_minosenter_y+fm_offset_y)*10.<<",       "<<(fm_minosenter_z+fm_offset_z)*10.<<", "<< tau <<")\n";
	      else
		myfile <<"      "<<fm_pdgcode<<"    "<<primindex-1<<"    0    (      "<<(fm_minosenter_x+fm_offset_x)*10.<<",       "<<(fm_minosenter_y+fm_offset_y)*10.<<",       "<<(fm_minosenter_z+fm_offset_z)*10.<<", "<< tau <<")\n";// 2.64e+06)\n";
	      
//          
        // std::cout<<fm_pdgcode<<" at z position "<<position.Z()<<" lost "<<((prevmomentum.E()-momentum.E())*1000.)<<"MeV in "<<(position.Z()-prevposition.Z())<<"cm which is "<<((prevmomentum.E()-momentum.E())*1000.)/(position.Z()-prevposition.Z())<< "MeV/cm VolumeName is " << geom->VolumeName(position.Vect())<<" "<<geom->MaterialName(position.Vect()) << std::endl;      

	      fm_energylost=fm_init_energy-fm_minosenter_energy;
	      fm_deflectionangle= TMath::ACos(((fm_tpcexit_px*fm_minosenter_px)+(fm_tpcexit_py*fm_minosenter_py)+(fm_tpcexit_pz*fm_minosenter_pz))/(sqrt(pow(fm_tpcexit_px,2)+pow(fm_tpcexit_py,2)+pow(fm_tpcexit_pz,2))*sqrt(pow(fm_minosenter_px,2)+pow(fm_minosenter_py,2)+pow(fm_minosenter_pz,2))));
//         std::cout<<"exit-enter "<<(fm_tpcexit_energy-fm_minosenter_energy)*1000.<<" "<<fm_deflectionangle<<" "<<fm_minosenter_energy<<std::endl;
            }
    }
    ftree->Fill();    
  }
  myfile << "________________________________________________________________________________\n";
  myfile.close(); 
  event++;
  return;  
}
  

  DEFINE_ART_MODULE(T962G4Ana)

} // namespace T962G4

#endif // T962G4_T962G4_H
